<?php
$data = <<<QUOTE
To be or not to be, that is the question;
Whether 'tis nobler in the mind to suffer
The Slings and Arrows of outrageous Fortune
Or to take arms against a sea of troubles,
And by opposing, end them. To die, to sleep;
No more; and by a sleep to say we end
The heart-ache and the thousand natural shocks
That flesh is heir to  'tis a consummation
Devoutly to be wish'd. To die, to sleep;
To sleep, perchance to dream. Ay, there's the rub,
For in that sleep of death what dreams may come,
When we have shuffled off this mortal coil,
Must give us pause. There's the respect
That makes calamity of so long life,
For who would bear the whips and scorns of time,
Th'oppressor's wrong, the proud man's contumely,
The pangs of dispriz'd love, the law's delay,
The insolence of office, and the spurns
That patient merit of th'unworthy takes,
When he himself might his quietus make
With a bare bodkin? who would fardels bear,
To grunt and sweat under a weary life,
But that the dread of something after death,
The undiscovered country from whose bourn
No traveller returns, puzzles the will,
And makes us rather bear those ills we have
Than fly to others that we know not of?
Thus conscience does make cowards of us all,
And thus the native hue of resolution
Is sicklied o'er with the pale cast of thought,
And enterprises of great pitch and moment
With this regard their currents turn away,
And lose the name of action.


Hath not a Jew eyes? Hath not a Jew hands, organs, dimensions, 
senses, affections, passions; fed with the same food, hurt with 
the same weapons, subject to the same diseases, heal'd by 
the same means, warm'd and cool'd by the same winter and summer 
as a Christian is? If you prick us, do we not bleed? If you 
tickle us, do we not laugh? If you poison us, do we not die? 
And if you wrong us, shall we not revenge? If we are like you 
in the rest, we will resemble you in that. If a Jew wrong a 
Christian, what is his humility? Revenge. If a Christian wrong 
a Jew, what should his sufferance be by Christian example? Why, 
revenge. The villainy you teach me, I will execute, and it 
shall go hard but I will better the instruction.

Is this a dagger which I see before me,
The handle toward my hand? Come, let me clutch thee.
I have thee not, and yet I see thee still.
Art thou not, fatal vision, sensible
To feeling as to sight? or art thou but
A dagger of the mind, a false creation,
Proceeding from the heat-oppress'd brain?
I see thee yet, in form as palpable
As this which now I draw.
Thou marshall'st me the way that I was going;
And such an instrument I was to use.
Mine eyes are made the fools o' the other senses,
Or else worth all the rest; I see thee still,
And on thy blade and dudgeon gouts of blood,
Which was not so before.
There's no such thing:
It is the bloody business which informs Thus to mine eyes. 
Now o'er the one halfworld Nature seems dead,
and wicked dreams abuse The curtain'd sleep; witchcraft celebrates
Pale Hecate's offerings, and wither'd murder,
Alarum'd by his sentinel, the wolf,
Whose howl's his watch, thus with his stealthy pace.
With Tarquin's ravishing strides, towards his design
Moves like a ghost. Thou sure and firm-set earth,
Hear not my steps, which way they walk, 
for fear Thy very stones prate of my whereabout,
And take the present horror from the time,
Which now suits with it.
Whiles I threat, he lives:
Words to the heat of deeds too cold breath gives.
I go, and it is done; the bell invites me.
Hear it not, Duncan; for it is a knell
That summons thee to heaven or to hell.

QUOTE;
